from pathlib import Path

import pytest

from kaizo import ConfigParser

X = 5

main_py = """
def raise_error():
  raise ValueError("error")
"""

ignore_config = """
local: main.py
x:
  module: local
  source: raise_error
  policy: ignore
"""

raise_config = """
local: main.py
x:
  module: local
  source: raise_error
  policy: raise
"""


def test_ignore_exception(tmp_path: Path) -> None:
    module = tmp_path / "main.py"
    module.write_text(main_py)

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(ignore_config)

    parser = ConfigParser(cfg_file)
    out = parser.parse()

    assert out["x"] is None


def test_raise_exception(tmp_path: Path) -> None:
    module = tmp_path / "main.py"
    module.write_text(main_py)

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(raise_config)

    parser = ConfigParser(cfg_file)
    out = parser.parse()

    with pytest.raises(ValueError, match="error"):
        out["x"]



from pathlib import Path

from kaizo import ConfigParser

X = 5

module_config = f"""
x: {X}
"""

relative_config = """
import:
  m: module.yml
run: m.{x}
"""

absolute_config = """
import:
  m: {tmp_path}/module.yml
run: m.{x}
"""


def test_relative_import_module(tmp_path: Path) -> None:
    module = tmp_path / "module.yml"
    module.write_text(module_config)

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(relative_config)

    parser = ConfigParser(cfg_file)
    out = parser.parse()

    assert out["run"] == X


def test_absolute_import_module(tmp_path: Path) -> None:
    module = tmp_path / "module.yml"
    module.write_text(module_config)

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(absolute_config.replace("{tmp_path}", str(tmp_path)))

    parser = ConfigParser(cfg_file)
    out = parser.parse()

    assert out["run"] == X



from dataclasses import dataclass

from typing_extensions import Self

from kaizo.utils.entry import DictEntry, ListEntry


@dataclass
class PluginMetadata:
    args: DictEntry[str] | ListEntry | None = None


class Plugin:
    metadata: PluginMetadata

    @classmethod
    def dispatch(cls: type[Self], metadata: PluginMetadata) -> Self:
        kwargs = {}
        args = ()

        if isinstance(metadata.args, DictEntry):
            kwargs = metadata.args
        elif isinstance(metadata.args, ListEntry):
            args = metadata.args

        plugin = cls(*args, **kwargs)
        plugin.metadata = metadata

        return plugin



import uuid
from abc import ABC, abstractmethod
from collections.abc import Generator, Iterable, MutableMapping, MutableSequence
from dataclasses import dataclass, field
from typing import Any, Generic, SupportsIndex, TypeVar

from typing_extensions import Self

from .cache import Cacheable
from .exception import ExceptionHandler, ExceptionPolicy
from .fn import FnWithKwargs

K = TypeVar("K")
V = TypeVar("V")


@dataclass
class Entry(ABC):
    key: str

    @abstractmethod
    def __call__(self) -> Any:
        pass


class DictEntry(MutableMapping, Cacheable, Generic[K]):
    _data: dict[K, Entry]
    _resolve: bool

    def __init__(self, data: dict[K, Entry] | None = None, *, resolve: bool = True) -> None:
        super().__init__()

        self._data = data or {}
        self._resolve = resolve

    @staticmethod
    def from_raw(
        root_key: str | None = None,
        raw_data: dict[K, Any] | None = None,
        *,
        resolve: bool = True,
    ) -> Self:
        if root_key is None:
            root_key = uuid.uuid4().hex

        if raw_data is None:
            raw_data = {}

        data = {
            key: FieldEntry(key=root_key, value=value) for key, value in raw_data.items()
        }

        return DictEntry(data=data, resolve=resolve)

    def __setitem__(self, key: K, value: Any) -> None:
        if not isinstance(value, Entry):
            msg = f"Value must be an Entry instance, got {type(value)}"
            raise TypeError(msg)

        self._data[key] = value
        self._update_id()

    def __getitem__(self, key: K) -> Any:
        value = self._data[key]

        if not isinstance(value, Entry):
            msg = f"Value must be an Entry instance, got {type(value)}"
            raise TypeError(msg)

        if self._resolve:
            return value.__call__()

        return value

    def __delitem__(self, key: K) -> None:
        self._data.__delitem__(key)
        self._update_id()

    def __iter__(self) -> Generator[K]:
        yield from self._data

    def __len__(self) -> int:
        return self._data.__len__()

    def __contains__(self, key: K) -> bool:
        return self._data.__contains__(key)


class ListEntry(MutableSequence, Cacheable):
    _data: list[Entry]
    _resolve: bool

    def __init__(self, data: list[Entry] | None = None, *, resolve: bool = True) -> None:
        super().__init__()

        self._data = data or []
        self._resolve = resolve

    @staticmethod
    def from_raw(
        root_key: str | None = None,
        raw_data: list[Any] | None = None,
        *,
        resolve: bool = True,
    ) -> Self:
        if root_key is None:
            root_key = uuid.uuid4().hex

        if raw_data is None:
            raw_data = []

        data = [FieldEntry(key=root_key, value=value) for value in raw_data]

        return ListEntry(data=data, resolve=resolve)

    def __setitem__(self, i: SupportsIndex, value: Any) -> None:
        if isinstance(value, Iterable):
            for value_i in value:
                if not isinstance(value_i, Entry):
                    msg = f"Value must be an Entry instance, got {type(value_i)}"
                    raise TypeError(msg)

        elif not isinstance(value, Entry):
            msg = f"Value must be an Entry instance, got {type(value)}"
            raise TypeError(msg)

        self._data[i] = value
        self._update_id()

    def __getitem__(self, i: SupportsIndex) -> Any:
        value = self._data.__getitem__(i)

        if isinstance(value, Iterable):
            new_values = []
            for value_i in value:
                if not isinstance(value_i, Entry):
                    msg = f"Value must be an Entry instance, got {type(value_i)}"
                    raise TypeError(msg)

                if self._resolve:
                    new_values.append(value_i.__call__())
                else:
                    new_values.append(value_i)

            return new_values

        if isinstance(value, Entry):
            if self._resolve:
                return value.__call__()

            return value

        msg = f"Value must be an Entry instance, got {type(value)}"
        raise TypeError(msg)

    def __delitem__(self, i: SupportsIndex) -> None:
        self._data.__delitem__(i)
        self._update_id()

    def __len__(self) -> int:
        return self._data.__len__()

    def insert(self, index: SupportsIndex, value: Any) -> None:
        if not isinstance(value, Entry):
            msg = f"Value must be an Entry instance, got {type(value)}"
            raise TypeError(msg)

        self._data.insert(index, value)
        self._update_id()


@dataclass
class FieldEntry(Entry, Generic[V]):
    value: V

    def __call__(self) -> V:
        return self.value


@dataclass
class ModuleEntry(Entry):
    obj: Any
    call: Any
    lazy: bool
    args: DictEntry[str] | ListEntry | None = None
    cache: bool = True
    policy: ExceptionPolicy = ExceptionPolicy.RAISE
    fn: FnWithKwargs = field(init=False)
    bucket: dict[str] = field(init=False)
    exception_handler: ExceptionHandler = field(init=False)

    def __post_init__(self) -> None:
        self.bucket = {}
        self.exception_handler = ExceptionHandler(policy=self.policy)

        if self.call is False:
            return

        kwargs = {}
        args = ()

        if isinstance(self.args, DictEntry):
            kwargs = self.args
        elif isinstance(self.args, ListEntry):
            args = self.args

        if self.call is True:
            if not callable(self.obj):
                msg = f"'{self.obj}' is not callable"
                raise TypeError(msg)

            self.fn = FnWithKwargs(fn=self.obj, args=args, kwargs=kwargs)
            return

        if not hasattr(self.obj, self.call):
            msg = f"'{self.obj}' has no attribute '{self.call}'"
            raise AttributeError(msg)

        fn = getattr(self.obj, self.call)

        if not callable(fn):
            msg = f"'{fn}' is not callable"
            raise TypeError(msg)

        self.fn = FnWithKwargs(fn=fn, args=args, kwargs=kwargs)

    def _call_fn(self) -> Any:
        with self.exception_handler:
            return self.fn.__call__()

    def __call__(self) -> Any | FnWithKwargs:
        if self.call is False:
            return self.obj

        if self.lazy:
            return self.fn

        if not self.cache:
            return self._call_fn()

        uid = None

        if self.args is not None:
            uid = self.args.uid

        if uid not in self.bucket:
            self.bucket[uid] = self._call_fn()

        return self.bucket[uid]



from pathlib import Path

from kaizo import ConfigParser
from kaizo.utils import FnWithKwargs

X = 5
Y = 6
Z = 7

main_py = f"""
def fn(x,y,z):
  return (x,y,z)

def fn2(cb):
    return cb({X},{Y})
"""


lazy_config = f"""
local: main.py
a:
    module: local
    source: fn
    lazy: true
    args:
        z: {Z}
b:
    module: local
    source: fn2
    args:
        cb: .{{a}}
"""


def test_lazy_injection(tmp_path: Path) -> None:
    module = tmp_path / "main.py"
    module.write_text(main_py)

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(lazy_config)

    parser = ConfigParser(cfg_file, kwargs={"injected": X})
    out = parser.parse()

    assert isinstance(out["a"], FnWithKwargs)
    assert out["b"] == (X, Y, Z)



from pathlib import Path

from kaizo import ConfigParser

X = 5

main_py = """
import math
def pow(x,y):
  return math.pow(x,y)
"""

simple_config = """
x: .{injected}
"""

kwargs_config = """
local: main.py
implicit:
  module: local
  source: pow
  call: true
  args:
    - .{injected}
    - 2
explicit:
  module: local
  source: pow
  call: true
  args:
    x: .{injected}
    y: 2
"""


def test_simple_injection(tmp_path: Path) -> None:
    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(simple_config)

    parser = ConfigParser(cfg_file, kwargs={"injected": X})
    out = parser.parse()
    assert out["x"] == X


def test_kwargs_injection(tmp_path: Path) -> None:
    module = tmp_path / "main.py"
    module.write_text(main_py)

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(kwargs_config)

    parser = ConfigParser(cfg_file, kwargs={"injected": X})
    out = parser.parse()
    assert out["implicit"] == X**2
    assert out["explicit"] == X**2



import importlib
from pathlib import Path

from .common import create_fake_plugin

X = 7
Y = 6
VAL = 2

correct_plugin_config = """
plugins:
  correct: MyPlugin
"""
correct_plugin_py = """
from kaizo import Plugin

class MyPlugin(Plugin):
    pass
"""

plugin_with_args_config = f"""
plugins:
  dummy:
    source: MyPlugin
    args:
      x: {X}
      y: {Y}
"""
plugin_with_args_py = """
from kaizo import Plugin

class MyPlugin(Plugin):
    def __init__(self, x,y):
        self.x = x
        self.y = y
"""

using_plugin_config = f"""
plugins:
  dummy: MyPlugin
fn:
  module: plugin
  source: dummy
  call: sqrt
  args:
    - {VAL}
"""
using_plugin_py = """
import math
from kaizo import Plugin

class MyPlugin(Plugin):
    def sqrt(self, num):
      return math.sqrt(num)
"""


def test_plugin_subclass(tmp_path: Path) -> None:
    create_fake_plugin(tmp_path, "correct", body=correct_plugin_py)
    kaizo = importlib.import_module("kaizo")

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(correct_plugin_config)

    parser = kaizo.ConfigParser(cfg_file)
    assert parser.plugins is not None
    assert "correct" in parser.plugins


def test_plugin_with_args(tmp_path: Path) -> None:
    create_fake_plugin(tmp_path, "dummy", body=plugin_with_args_py)
    kaizo = importlib.import_module("kaizo")

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(plugin_with_args_config)

    parser = kaizo.ConfigParser(cfg_file)
    obj = parser.plugins["dummy"]

    assert isinstance(obj, kaizo.utils.FnWithKwargs)

    plugin = obj()
    assert plugin.x == X
    assert plugin.y == Y


def test_using_plugin(tmp_path: Path) -> None:
    create_fake_plugin(tmp_path, "dummy", body=using_plugin_py)
    kaizo = importlib.import_module("kaizo")

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(using_plugin_config)

    parser = kaizo.ConfigParser(cfg_file)

    out = parser.parse()
    entry = out["fn"]

    assert entry == VAL**0.5



from .cache import Cacheable
from .common import extract_variable
from .entry import DictEntry, Entry, FieldEntry, ListEntry, ModuleEntry
from .exception import ExceptionHandler, ExceptionPolicy
from .fn import FnWithKwargs
from .module import ModuleLoader
from .storage import Storage

__all__ = (
    "Cacheable",
    "DictEntry",
    "Entry",
    "ExceptionHandler",
    "ExceptionPolicy",
    "FieldEntry",
    "FnWithKwargs",
    "ListEntry",
    "ModuleEntry",
    "ModuleLoader",
    "Storage",
    "extract_variable",
)



import importlib
from pathlib import Path

import pytest

from kaizo import ConfigParser
from kaizo.utils import FnWithKwargs

VAL = 9

false_call_config = """
math_sqrt:
  module: math
  source: sqrt
  call: false
"""

not_callable_py = """
x = 2
"""
not_callable_config = """
local: main.py
not_callable:
  module: local
  source: x
"""

bad_attribute_config = """
math_sqrt:
  module: math
  source: bad_attribute
"""

not_callable_explicit_py = """
class Dummy:
  x = 2
"""
not_callable_explicit_config = """
local: main.py
not_callable:
  module: local
  source: Dummy
  call: x
"""

lazy_call_config = f"""
fn:
  module: math
  source: sqrt
  lazy: true
  args:
    - {VAL}
"""

correct_import_args_config = f"""
args:
 - {VAL}
fn:
  module: math
  source: sqrt
  args: .{{args}}
"""

incorrect_import_args_config = """
fn:
  module: math
  source: sqrt
  args: not_correct
"""

invalid_args_config = """
fn:
  module: math
  source: sqrt
  args: 1
"""


def test_false_call(tmp_path: Path) -> None:
    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(false_call_config)

    math = importlib.import_module("math")

    parser = ConfigParser(cfg_file)
    out = parser.parse()

    entry = out["math_sqrt"]

    assert entry is math.sqrt


def test_not_callable(tmp_path: Path) -> None:
    module = tmp_path / "main.py"
    module.write_text(not_callable_py)

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(not_callable_config)

    parser = ConfigParser(cfg_file)

    with pytest.raises(TypeError):
        parser.parse()


def test_bad_attribute(tmp_path: Path) -> None:
    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(bad_attribute_config)

    parser = ConfigParser(cfg_file)

    with pytest.raises(AttributeError):
        parser.parse()


def test_not_callable_explicit(tmp_path: Path) -> None:
    module = tmp_path / "main.py"
    module.write_text(not_callable_explicit_py)

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(not_callable_explicit_config)

    parser = ConfigParser(cfg_file)

    with pytest.raises(TypeError):
        parser.parse()


def test_lazy_call(tmp_path: Path) -> None:
    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(lazy_call_config)

    parser = ConfigParser(cfg_file)
    out = parser.parse()

    entry = out["fn"]
    assert isinstance(entry, FnWithKwargs)

    assert entry() == VAL**0.5


def test_correct_import_args(tmp_path: Path) -> None:
    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(correct_import_args_config)

    parser = ConfigParser(cfg_file)
    out = parser.parse()

    entry = out["fn"]

    assert entry == VAL**0.5


def test_incorrect_import_args(tmp_path: Path) -> None:
    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(incorrect_import_args_config)

    parser = ConfigParser(cfg_file)

    with pytest.raises(TypeError, match=r"args must be `ListEntry` or `DictEntry`, got *"):
        parser.parse()


def test_invalid_args_type(tmp_path: Path) -> None:
    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(invalid_args_config)

    parser = ConfigParser(cfg_file)

    with pytest.raises(TypeError, match=r"invalid type for args, got *"):
        parser.parse()



from pathlib import Path

from kaizo import ConfigParser

VAL = 16

config = f"""
square:
  module: math
  source: sqrt
  call: true
  args:
    - {VAL}
"""


def test_module_resolution(tmp_path: Path) -> None:
    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(config)

    parser = ConfigParser(cfg_file)
    out = parser.parse()
    entry = out["square"]

    assert entry == VAL**0.5



import re
from enum import Enum


class StrEnum(str, Enum):
    def __str__(self) -> str:
        return self.value


def extract_variable(entry: str) -> tuple[str | None, str | None, str]:
    matched = re.match(r"^(\w*)\.(?:(\w+)\.)?\{(\w*)\}$", entry)
    if matched is None:
        return None, None, entry

    return matched.groups()



from collections.abc import Callable
from functools import partial
from typing import Generic, TypeVar

R = TypeVar("R")


class FnWithKwargs(Generic[R]):
    fn: Callable[..., R]
    args: tuple
    kwargs: dict[str]

    def __init__(
        self,
        fn: Callable[..., R],
        args: tuple | None = None,
        kwargs: dict[str] | None = None,
    ) -> None:
        if args is None:
            args = ()

        if kwargs is None:
            kwargs = {}

        self.fn = fn
        self.args = args
        self.kwargs = kwargs

    def __call__(self, *args, **kwargs) -> R:
        fn = partial(self.fn, *self.args, **self.kwargs)

        return fn(*args, **kwargs)

    def update(self, **kwargs) -> None:
        self.kwargs.update(kwargs)



from collections.abc import Callable, Iterable
from contextlib import AbstractContextManager
from types import TracebackType

from typing_extensions import Self

from .common import StrEnum


class ExceptionPolicy(StrEnum):
    RAISE = "raise"
    IGNORE = "ignore"


class ExceptionHandler(AbstractContextManager):
    _policy: ExceptionPolicy
    _exc_types: tuple[type[BaseException]]
    _exc_handler: Callable[[], bool]

    def __init__(
        self,
        policy: ExceptionPolicy,
        exc_types: Iterable[type[BaseException]] = (Exception,),
    ) -> None:
        self._policy = policy
        self._exc_types = tuple(exc_types)
        self._exc_handler = None

    def __enter__(self) -> Self:
        match self._policy:
            case ExceptionPolicy.RAISE:
                self._exc_handler = self._raise
            case ExceptionPolicy.IGNORE:
                self._exc_handler = self._ignore

        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        if exc_type is None:
            return False

        if not isinstance(exc_value, self._exc_types):
            return False

        return self._exc_handler()

    def _raise(self) -> bool:
        return False

    def _ignore(self) -> bool:
        return True
